/*
 * xroottext.c
 * version 0.7     March 9, 1999
 * Copyright (c) 1999 Robert Forsman
 * <thoth@purplefrog.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/*
 * Paints text from stdin into a primitive pseudo-tty rendered onto
 * the root window.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>

/**********************************************************************/

char *progname;

/**********************************************************************/

void usage()
{
    fprintf(stderr, "Usage:\n  %s [ -v ] [ -display dpyname ] [ -geom WxH+X+Y ] [ -fn fontspec ] [ -fg color ] [ -bg color ] < text\n", progname);
}

struct pseudo_tty {
    int	ncols, nrows;
    int	xoff, yoff;
    char *textbuf;		/* array of ncols*nrows characters */
    int	cursor_row, cursor_col;
    int scrolled_since_display;

    int	demi_opaque;		/* use XDrawImageString instead of
				   XDrawString */
};

void init_pseudo_tty(struct pseudo_tty *that,
		     int nrows, int ncols, int xoff, int yoff,
		     int demi_opaque)
{
    that->nrows = nrows;
    that->ncols = ncols;
    that->xoff = xoff;
    that->yoff = yoff;
    that->demi_opaque = demi_opaque;

    that->cursor_row = 0;
    that->cursor_col = 0;
    that->scrolled_since_display = 0;
    that->textbuf = (char*)malloc(nrows*ncols);
    memset(that->textbuf, 0, nrows*ncols);
}

void scroll_pty_1line(struct pseudo_tty *pty)
{
    int row, col;
    /* scroll lines up one */
    memcpy(pty->textbuf, pty->textbuf + pty->ncols, pty->ncols*(pty->nrows-1));
    /* NUL out last line */
    memset(pty->textbuf + (pty->ncols*(pty->nrows-1)), 0, pty->ncols);
    pty->scrolled_since_display = 1;
}

void pseudo_tty_newline(struct pseudo_tty *pty)
{
    if (pty->cursor_row+1 >= pty->nrows) {
	scroll_pty_1line(pty);
	pty->cursor_row = pty->nrows-1;
    } else {
	pty->cursor_row++;
    }
    pty->cursor_col = 0;
}

void pseudo_tty_putchar(char c, struct pseudo_tty *pty)
{
    if (c == '\n') {
	pseudo_tty_newline(pty);
    } else if (c == '\t') {
	/* this breaks near the end of line */
	int endcol = (pty->cursor_col / 8 +1) *8;
	if (endcol>=pty->ncols) {
	    pseudo_tty_newline(pty);
	} else {
	    while (pty->cursor_col<endcol)
		pseudo_tty_putchar(' ', pty);
	}
    } else {
	if (c == 0) {
	    c = ' ';		/* I could revamp the rendering engine
				   to cope with this, but I'm lazy */
	}
	pty->textbuf[ pty->cursor_row*pty->ncols + pty->cursor_col ] = c;
	if (pty->cursor_col+1 >= pty->ncols) {
	    pseudo_tty_newline(pty);
	} else {
	    pty->cursor_col++;
	}
    }
}


/**********************************************************************/

int verbose = 0;

/**********************************************************************/

void redraw_pseudo_tty(struct pseudo_tty *pty, Display *dpy, Window win, GC gc, XFontStruct *font)
{
    int	row, col;
    int	str_width, ascent, height;
    int	tty_pixwidth;
    char	*buf;
    int	y;

    buf = malloc(pty->ncols+1);	/* make room for a NUL */
    buf[pty->ncols] = 0;

    ascent = font->ascent;
    height = ascent + font->descent;
    tty_pixwidth = font->max_bounds.width * pty->ncols;

    for (row=0; row < pty->nrows; row++) {
	memcpy(buf, pty->textbuf + (row*pty->ncols), pty->ncols);

	str_width = XTextWidth(font, buf, strlen(buf));
	y = pty->yoff+row*height;

	if (pty->demi_opaque) {
	    XClearArea(dpy, win, pty->xoff+str_width, y,
		       tty_pixwidth - str_width, height, 0);
	    if (buf[0]!=0)
	    XDrawImageString(dpy, win, gc, pty->xoff, y+ascent, buf, strlen(buf));
	} else {
	    XClearArea(dpy, win, pty->xoff, y,
		       tty_pixwidth, height, 0);
	    if (buf[0]!=0)
		XDrawString(dpy, win, gc, pty->xoff, y+ascent, buf, strlen(buf));
	    if (verbose>2)
		fprintf(stderr, "rendered row %d @%d+%d `%s'\n", row, pty->xoff, y+ascent, buf);
	}
    }
}

void erase_pseudo_tty(struct pseudo_tty *pty, Display *dpy, Window win, GC gc,
		      XFontStruct *font)
{
    int	ch, cw;
    ch = font->ascent + font->descent;
    cw = font->max_bounds.width;

    XClearArea(dpy, win, pty->xoff, pty->yoff,
	       cw*pty->ncols, ch*pty->nrows, 1);
}

/**********************************************************************/

/* The below function was modified from the xsetroot source */
unsigned long
NameToPixel(Display *dpy, int screen, char * name, unsigned long pixel)
{
    XColor ecolor;

    if (!name || !*name)
        return pixel;
    if (!XParseColor(dpy,DefaultColormap(dpy,screen),name,&ecolor)) {
        fprintf(stderr,"%s:  unknown color \"%s\"\n","doody",name);
        exit(1);
        /*NOTREACHED*/
    }
    if (!XAllocColor(dpy, DefaultColormap(dpy, screen),&ecolor)) {
        fprintf(stderr, "%s:  unable to allocate color for \"%s\"\n",
                "doody", name);
        exit(1);
        /*NOTREACHED*/
    }
    return(ecolor.pixel);
}

/**********************************************************************/

int expose_intersects_pseudo_tty(XExposeEvent *ev, struct pseudo_tty *pty, XFontStruct *font)
{
    return 1;			/* stub */
}

void devour_subsequent_expose_events(Display *dpy)
{
    XEvent ev;
    while (XPending(dpy)) {
	XPeekEvent(dpy, &ev);
	if (ev.type == Expose) {
	    XNextEvent(dpy, &ev);
	} else {
	    break;
	}
    }
}

void do_X_events(Display *dpy, Window win, GC gc, XFontStruct *font,
		 struct pseudo_tty *pty)
{
    XEvent ev;

    if (!XPending(dpy))
	return;

    XNextEvent(dpy, &ev);
    if (ev.type == Expose) {
	XExposeEvent	*evx = &ev.xexpose;
#if 1
	devour_subsequent_expose_events(dpy);
#else
	if (expose_intersects_pseudo_tty(&ev.xexpose, pty, font))
#endif
	    {
	    redraw_pseudo_tty(pty, dpy, win, gc, font);
	}
    } else {
	fprintf(stderr, "%s: unexpected event type %d\n",
		progname, ev.type);
    }
}

/**********************************************************************/

Window root;
Display *dpy;
int screen;
GC gc;
XFontStruct *font;
int	font_height, font_width;
unsigned long	fg_pixel;

/**********************************************************************/

int forever = 1;

void interrupted(int x)
{
    forever = 0;
}

void input_loop(struct pseudo_tty *pty)
{
    int stdin_open = 1;
    int	Xfd = ConnectionNumber(dpy);

    fd_set	readfds;

    while (forever) {
	XFlush(dpy);

	FD_ZERO(&readfds);
	if(stdin_open)
	    FD_SET(0, &readfds);
	FD_SET(Xfd, &readfds);

	if (select(Xfd+1, &readfds, (fd_set*)0, (fd_set*)0, (struct timeval*)0) <0) {
	    if (errno==EINTR) {
		/* whatever */
	    } else {
		fprintf(stderr, "%s: error during select ");
		perror("");
		exit(1);
	    }
	}

	if (FD_ISSET(0, &readfds)) {
	    char	buf[4096];
	    int	rval;
	    rval = read(0, buf, sizeof(buf));
	    if (rval<0) {
		if (errno == EINTR) {
		    /* like it's a big deal */
		} else {
		    fprintf(stderr, "%s: error during read ");
		    perror("");
		    stdin_open = 0;
		}
	    } else if (rval ==0) {
		stdin_open = 0;
	    } else {
		int	i;
		for (i=0; i<rval; i++) {
		    pseudo_tty_putchar(buf[i], pty);
		}
		redraw_pseudo_tty(pty, dpy, root, gc, font);
	    }
	}

	do_X_events(dpy, root, gc, font, pty);
    }

}

/**********************************************************************/
/*                                                                    */
/**********************************************************************/

int main(int argc, char **argv)

{
    int ncols = 80, nrows = 25;
    int xoff = 5, yoff = 5;		/* totally arbitrary defaults */

    int i;
    char *dpyname = 0;
    char *fontname = "*";
    char *geomstring=0;
    char *fgname=0;
    char *bgname=0;
    struct pseudo_tty	pty;


    progname = argv[0];

    for (i=1; i<argc; i++) {
	if (0==strcmp(argv[i], "-v")) {
	    verbose++;
	} else if (0==strcmp(argv[i], "-display")) {
	    if (i+1 >= argc) {
		fprintf(stderr, "%s: -display requires a display name for an argument\n", progname);
		usage();
		exit(1);
	    }

	    dpyname = argv[++i];
	} else if (0==strcmp(argv[i], "-fn")) {
	    if (i+1 >= argc) {
		fprintf(stderr, "%s: -fn requires a font name for an argument\n", progname);
		usage();
		exit(1);
	    }

	    fontname = argv[++i];
	} else if (0==strcmp(argv[i], "-geom")) {
	    if (i+1 >= argc) {
		fprintf(stderr, "%s: -geom requires a geometry spec for an argument\n", progname);
		usage();
		exit(1);
	    }

	    geomstring = argv[++i];
	} else if (0==strcmp(argv[i], "-fg")) {
	    if (i+1 >= argc) {
		fprintf(stderr, "%s: -fg requires a color spec for an argument\n", progname);
		usage();
		exit(1);
	    }

	    fgname = argv[++i];
	} else if (0==strcmp(argv[i], "-bg")) {
	    if (i+1 >= argc) {
		fprintf(stderr, "%s: -bg requires a color spec for an argument\n", progname);
		usage();
		exit(1);
	    }

	    bgname = argv[++i];
	} else {
	    fprintf(stderr, "%s: unknown flag %s\n", progname, argv[i]);
	    usage();
	    exit(1);
	}
    }

    if (!(dpy = XOpenDisplay(dpyname))) {
	fprintf(stderr, "%s: Unable to open display %s\n", dpyname);
	exit(1);
    }

    screen=DefaultScreen(dpy);

    if ((font = XLoadQueryFont(dpy, fontname))==NULL) {
	fprintf(stderr, "%s: Unable to load font %s\n", fontname);
	exit(1);
    }

    font_height = font->ascent + font->descent;
    font_width = font->max_bounds.width;

    if (geomstring != 0) {
	int	x,y;
	unsigned int	w,h;
	int	flags;

	flags = XParseGeometry(geomstring, &x, &y, &w, &h);
	if (verbose>1)
	  printf("parsed geometry string '%s' got %d\n", geomstring, flags);
	if (flags&WidthValue) {
	    ncols = w;
	}
	if (flags&HeightValue) {
	    nrows = h;
	}
	if (flags &XValue) {
	    if (flags &XNegative) {
		xoff = DisplayWidth(dpy, screen) + x - (font_width * ncols);
	    } else {
		xoff = x;
	    }
	}
	if (flags &YValue) {
	    if (flags &YNegative) {
		yoff = DisplayHeight(dpy, screen) + y - (font_height * nrows);
	    } else {
		yoff = y;
	    }
	}
    }

    if (verbose>1) {
	fprintf(stderr, "%s: using window %dx%d+%d+%d\n", progname,
		font_width*ncols, font_height*nrows,
		xoff, yoff);
    }

    init_pseudo_tty(&pty, nrows, ncols, xoff, yoff, (bgname!=0));

    root=RootWindow(dpy, screen);
    gc = DefaultGC(dpy, screen);

    XSetFont(dpy, gc, font->fid);

    fg_pixel = WhitePixel(dpy, screen);
    if (fgname != 0)
	fg_pixel = NameToPixel(dpy, screen, fgname, fg_pixel);

    XSetForeground(dpy, gc, fg_pixel);

    if (bgname != 0) {
	unsigned long bg_pixel = NameToPixel(dpy, screen, bgname, BlackPixel(dpy, screen));

	XSetBackground(dpy, gc, bg_pixel);
    }

    XSelectInput(dpy, root, ExposureMask);

    {
	struct sigaction	act;
	act.sa_handler = interrupted;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	sigaction(SIGHUP, &act, (struct sigaction*)0);
	sigaction(SIGINT, &act, (struct sigaction*)0);
	sigaction(SIGQUIT, &act, (struct sigaction*)0);
	sigaction(SIGTERM, &act, (struct sigaction*)0);
    }

    if (verbose>0) {
	fprintf(stderr, "xroottext v0.8\nCopyright 1999 Robert Forsman <thoth@purplefrog.com>\n");
	fprintf(stderr, "\
%s comes with ABSOLUTELY NO WARRANTY;\n\
This is free software, and you are welcome to redistribute it\n\
under the GNU General Public License as published by the Free\n\
Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.\n", "xroottext");

    }

    input_loop(&pty);

    if (verbose>1)
	fprintf(stderr, "erasing\n");
    erase_pseudo_tty(&pty, dpy, root, gc, font);

    XCloseDisplay(dpy);
}
